#include <windows.h>
#include <winioctl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strsafe.h>

#define DRIVER_INSTALLER
#include <config.h>
#include "common/debug.h"

#define IOCTL_SETBDF 60

#define IOCTL_GET_LAST_BRANCH_IO     \
        CTL_CODE(IOCTL_SETBDF,        \
                 0x0, \
				 METHOD_BUFFERED, \
                 FILE_ANY_ACCESS )

#define CURRENT_DIR_BUF_LEN 500

static int
InstallDriver(
    __in SC_HANDLE  SchSCManager,
    __in LPCTSTR    DriverName,
    __in LPCTSTR    ServiceExe )
{
	SC_HANDLE   schService;
	DWORD       err;

	schService = CreateService ( SchSCManager,         // handle of service control manager database
							   DriverName,             // address of name of service to start
							   DriverName,             // address of display name
							   SERVICE_ALL_ACCESS,     // type of access to service
							   SERVICE_KERNEL_DRIVER,  // type of service
							   SERVICE_DEMAND_START,   // when to start service
							   SERVICE_ERROR_NORMAL,   // severity if service fails to start
							   ServiceExe,             // address of name of binary file
							   NULL,                   // service does not belong to a group
							   NULL,                   // no tag requested
							   NULL,                   // no dependency names
							   NULL,                   // use LocalSystem account
							   NULL                    // no password for service account
							   );

	if ( schService == NULL ) 
	{
		err = GetLastError();
		if (err == ERROR_SERVICE_EXISTS) 
		{
			// Ignore this error.
			LogWrn("service exists!\n");
			return 0;
		} else 
		{
			LogErr("CreateService failed!  Error = %d \n", err );
			// Indicate an error.
			return  -1;
		}
	}
	
	// Close the service object.
	if (schService) 
	{
		CloseServiceHandle(schService);
	}
	
	// Indicate success.
	return 0;
}   // InstallDriver

static int StartDriver( __in SC_HANDLE SchSCManager, __in LPCTSTR DriverName, __out HANDLE * FileHandle)
{
    SC_HANDLE   schService;
    DWORD       err;
    int	status = 0;
	TCHAR createDriverName[CURRENT_DIR_BUF_LEN];

    // Open the handle to the existing service.
    schService = OpenService( SchSCManager, DriverName, SERVICE_ALL_ACCESS );
    if ( schService == NULL )
	{
        LogErr( "OpenService failed!  Error = %d \n", GetLastError() );
        // Indicate failure.
        return -1;
    }

    // Start the execution of the service (i.e. start the driver).
    if ( !StartService( schService,     // service identifier
                      0,                // number of arguments
                      NULL              // pointer to arguments
                      ) ) 
	{
        err = GetLastError();
        if ( err == ERROR_SERVICE_ALREADY_RUNNING ) 
		{
            // Ignore this error.
			LogWrn( "Service has been alread run!\n" );
			status = 0;
        } 
		else 
		{
            LogErr( "StartService failure! Error = %d \n", err );
            // Indicate failure.  Fall through to properly close the service handle.
			status = -1;
        }
    }
    
	// Close the service object.
    if ( schService )
	{
        CloseServiceHandle( schService );
    }

	if( FileHandle )
	{
		if ( ( StringCbCopy( createDriverName, CURRENT_DIR_BUF_LEN, "\\\\.\\" ) ) ) 
		{
			LogErr("StringCBCat error %d!\n", GetLastError());
			return -1;
		}
	
		if ( ( StringCbCat( createDriverName, CURRENT_DIR_BUF_LEN, DriverName ) ) ) 
		{
			LogErr("StringCBCat error %d!\n", GetLastError());
			return -1;
		}

		*FileHandle = CreateFile( createDriverName,
                    GENERIC_READ | GENERIC_WRITE,
                    0,
                    NULL,
                    CREATE_ALWAYS,
                    FILE_ATTRIBUTE_NORMAL,
                    NULL);

		if ( *FileHandle == INVALID_HANDLE_VALUE ) 
		{
			LogErr( "%s is not loaded.\n", DriverName );
			// Indicate failure.
			return -1;
		}
	}

    return status;
}   // StartDriver

static int StopDriver( __in SC_HANDLE SchSCManager, __in LPCTSTR DriverName, HANDLE fileHandle )
{
    int         rCode = 0;
    SC_HANDLE       schService;
    SERVICE_STATUS  serviceStatus;

	// Close open file handle.
	if ( fileHandle )
	{
		CloseHandle( fileHandle );
	}

    // Open the handle to the existing service.
    schService = OpenService( SchSCManager,
                             DriverName,
                             SERVICE_ALL_ACCESS
                             );
    if ( !schService ) 
	{
		LogErr( "OpenService failed! Error = %d \n", GetLastError() );
		return -1;
    }

    // Request that the service stop.
    if ( ControlService( schService, SERVICE_CONTROL_STOP, &serviceStatus ) ) 
	{
        // Indicate success.
        rCode = 0;
    }
	else 
	{
		LogErr( "ControlService failed! Can't stop driver. Error = %d \n", GetLastError() );
		// Indicate failure. Fall through to properly close the service handle.
		rCode = -1;
    }

    // Close the service object.
    if ( schService ) 
        CloseServiceHandle ( schService );

    return rCode;
}   //  StopDriver

static int RemoveDriver( __in SC_HANDLE SchSCManager, __in LPCTSTR DriverName )
{
    SC_HANDLE   schService;
    int     rCode = 0;

    // Open the handle to the existing service.
    schService = OpenService( SchSCManager, DriverName, SERVICE_ALL_ACCESS );
    if ( schService == NULL ) 
	{
        printf( "OpenService failed!  Error = %d \n", GetLastError() );
        // Indicate error.
        return -1;
    }

    // Mark the service for deletion from the service control manager database.
    if ( DeleteService( schService ) ) 
	{
        // Indicate success.
        rCode = 0;
    }
	else 
	{
        LogErr( "DeleteService failed!  Error = %d \n", GetLastError() );
        // Indicate failure.  Fall through to properly close the service handle.
        rCode = -1;
    }

    // Close the service object.
    if ( schService )
        CloseServiceHandle( schService );

    return rCode;
}   // RemoveDriver

int LoadDriver( char * driverName, HANDLE * fileHandle )
{
	DWORD errNum = 0, driverLocLen = 0;
	void *res = NULL;
	unsigned long resLen = 0, unpackedSize = 0;
	int retVal = 0;
	size_t destLen = 0;
	unsigned char * dest = NULL;
	SC_HANDLE   schSCManager;
	BOOLEAN rCode = TRUE;
	OSVERSIONINFO info={0};
	
	TCHAR fullDriverName[CURRENT_DIR_BUF_LEN];

	driverLocLen = GetCurrentDirectory( CURRENT_DIR_BUF_LEN, fullDriverName );
	if ( !driverLocLen )
	{
		LogErr("Can't get current directory! err %d \n", GetLastError() );
		return -1;
	}

	if ( ( StringCbCat( fullDriverName, CURRENT_DIR_BUF_LEN, "\\" ) ) ) 
	{
		LogErr("StringCBCat error %d!\n", GetLastError());
        return -1;
	}

    if ( ( StringCbCat( fullDriverName, CURRENT_DIR_BUF_LEN, driverName ) ) ) 
	{
		LogErr("StringCBCat error %d!\n", GetLastError());
        return -1;
	}

	if ( ( StringCbCat( fullDriverName, CURRENT_DIR_BUF_LEN, ".sys" ) ) ) 
	{
		LogErr("StringCBCat error %d!\n", GetLastError());
        return -1;
	}

	LogDbg( "Full driver path is %s\n", fullDriverName );

	// Connect to the Service Control Manager and open the Services database.
	schSCManager = OpenSCManager(NULL,                   // local machine
								 NULL,                   // local database
								 SC_MANAGER_ALL_ACCESS   // access required
								 );
	if (!schSCManager) 
	{
		LogErr("Open SC Manager failed! Error = %d \n", GetLastError());
		return -1;;
	}
	if ( InstallDriver( schSCManager, driverName, fullDriverName ) ) 
	{
		LogErr("Can't install this driver!\n");
		CloseServiceHandle(schSCManager);
		return -1;
	} 

	// Start the driver service (i.e. start the driver).
	LogDbg("Driver is installed\n");
	if ( StartDriver( schSCManager, driverName, fileHandle ) )
	{
		LogErr("Can't start this driver\n");
		if( RemoveDriver( schSCManager, driverName ) )
		{
			LogErr( "Can't remove driver\n" );
		}
		CloseServiceHandle(schSCManager);
		return -1;
	}

	LogDbg("Driver is started\n");
	CloseServiceHandle(schSCManager);	
	
	return 0;
}

int UnloadDriver( char * driverName, HANDLE fileHandle )
{
	SC_HANDLE   schSCManager;
	TCHAR fullDriverName[CURRENT_DIR_BUF_LEN];
	DWORD driverLocLen = 0;
	
	driverLocLen = GetCurrentDirectory( CURRENT_DIR_BUF_LEN, fullDriverName );
	if ( !driverLocLen )
	{
		LogErr("Can't get current directory! err %d \n", GetLastError() );
		return -1;
	}
	
	if ( ( StringCbCat( fullDriverName, CURRENT_DIR_BUF_LEN, "\\" ) ) ) 
	{
		LogErr("StringCBCat error %d!\n", GetLastError());
        return -1;
	}
	
    if ( ( StringCbCat( fullDriverName, CURRENT_DIR_BUF_LEN, driverName ) ) ) 
	{
		LogErr("StringCBCat error %d!\n", GetLastError());
        return -1;
	}
	
	if ( ( StringCbCat( fullDriverName, CURRENT_DIR_BUF_LEN, ".sys" ) ) ) 
	{
		LogErr("StringCBCat error %d!\n", GetLastError());
        return -1;
	}

	// Connect to the Service Control Manager and open the Services database.
	schSCManager = OpenSCManager(NULL,                   // local machine
								 NULL,                   // local database
								 SC_MANAGER_ALL_ACCESS   // access required
								 );
	if (!schSCManager) 
	{
		LogErr("Open SC Manager failed! Error = %d \n", GetLastError());
		return -1;;
	}

	if ( StopDriver( schSCManager, driverName, fileHandle ) ) 
	{
		LogErr("Can't stop this driver!\n");
	}
	else
	{
		LogDbg("Driver is stoped\n");
	}

	if ( RemoveDriver( schSCManager, driverName ) )
	{
		LogErr("Can't remove this driver\n");
	} 
	else
	{
		LogDbg( "Driver is removed\n" );
	}
	
	CloseServiceHandle(schSCManager);	

	return 0;
}


int GetBranchIOCTLRequest( HANDLE hDevice, DWORD Eip, DWORD * lastBranchEip )
{
	/*HANDLE hDevice;*/
	DWORD bytesReturned = 0;
	BOOLEAN bRc = FALSE;

    bRc = DeviceIoControl ( hDevice,
                            (DWORD) IOCTL_GET_LAST_BRANCH_IO, 
                            &Eip,
                            sizeof ( DWORD ),
                            lastBranchEip,
                            sizeof( DWORD),
                            &bytesReturned,
                            NULL
                            );

    if ( !bRc )
    {
        printf ( "Error in DeviceIoControl : %d", GetLastError());
        return -1;
    }

	return 0;
}